@@ -22,13 +22,13 @@ import logging
 import logging.config
 import multiprocessing
 import os
-import subprocess
 from os.path import abspath, basename, dirname, expanduser, exists, isfile, join, splitext
 # from pprint import pprint, pformat
 import platform
 import plistlib
 import pstats
-# import subprocess
+import shlex
+import subprocess
 import sys
 import tempfile
 import threading
@@ -46,6 +46,36 @@ MAIN_CFG_FILENAME = 'app.json'
 DEFAULT_CFG_FILENAME = 'default.json'
 
 
+class SingletonDecorator:
+    """
+    Decorator to build Singleton class, single-inheritance only.
+    Usage:
+        class MyClass: ...
+        myobj = SingletonDecorator(MyClass, args, kwargs)
+    """
+    def __init__(self, klass):
+        self.klass = klass
+        self.instance = None
+
+    def __call__(self, *args, **kwargs):
+        if self.instance is None:
+            self.instance = self.klass(*args, **kwargs)
+        return self.instance
+
+
+class _Singleton(type):
+    _instances = {}
+
+    def __call__(cls, *args, **kwargs):
+        if cls not in cls._instances:
+            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
+        return cls._instances[cls]
+
+
+class Singleton(_Singleton('SingletonMeta', (object,), {})):
+    pass
+
+
 class LowPassLogFilter(object):
     """
     Logging filter: Show log messages below input level.
@@ -690,7 +720,9 @@ def get_clipboard_content():
 def alert(title, content, action='Close'):
     if platform.system() == 'Windows':
         cmd = ['mshta', f'vbscript:Execute("msgbox ""{content}"", 0,""{title}"":{action}")']
-    elif platform.system() == 'Darwin':
+        os.system(' '.join(cmd))
+        return
+    if platform.system() == 'Darwin':
         cmd = ['osascript', '-e', f'display alert "{title}" message "{content}"']
     else:
         cmd = ['echo', f'{title}: {content}: {action}']
@@ -796,39 +828,90 @@ def rerun_lock(name, folder=None, infohook=_logger.info, warnhook=_logger.warnin
     return decorator
 
 
-class SingletonDecorator:
+def append_to_os_paths(bindir):
     """
-    Decorator to build Singleton class, single-inheritance only.
-    Usage:
-        class MyClass: ...
-        myobj = SingletonDecorator(MyClass, args, kwargs)
+    On macOS, PATH update will only take effect after calling `source ~/.bash_profile` directly in shell. It won't work 
     """
-    def __init__(self, klass):
-        self.klass = klass
-        self.instance = None
-
-    def __call__(self, *args, **kwargs):
-        if self.instance is None:
-            self.instance = self.klass(*args, **kwargs)
-        return self.instance
+    if platform.system() == 'Windows':
+        import winreg
+        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as reg:
+            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Environment', 0, winreg.KEY_ALL_ACCESS) as key:
+                user_paths, _ = winreg.QueryValueEx(key, 'Path')
+                if bindir not in user_paths:
+                    if user_paths[-1] != ';':
+                        user_paths += ';'
+                    user_paths += f'{bindir}'
+                    winreg.SetValueEx(key, 'Path', 0, winreg.REG_EXPAND_SZ, user_paths)
+    else:
+        cfg_file = os.path.expanduser('~/.bash_profile') if platform.system() == 'Darwin' else os.path.expanduser('~/.bashrc')
+        if bindir in os.environ['PATH']:
+            return
+        with open(cfg_file, 'a') as fp:
+            fp.write(f'\nexport PATH="$PATH:{bindir}"\n\n')
+    os.environ['PATH'] += os.pathsep + bindir
 
 
-class _Singleton(type):
-    _instances = {}
-
-    def __call__(cls, *args, **kwargs):
-        if cls not in cls._instances:
-            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
-        return cls._instances[cls]
+def prepend_to_os_paths(bindir):
+    if platform.system() == 'Windows':
+        import winreg
+        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as reg:
+            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Environment', 0, winreg.KEY_ALL_ACCESS) as key:
+                user_paths, _ = winreg.QueryValueEx(key, 'Path')
+                if bindir not in user_paths:
+                    user_paths = f'{bindir};' + user_paths
+                    winreg.SetValueEx(key, 'Path', 0, winreg.REG_EXPAND_SZ, user_paths)
+    else:
+        cfg_file = os.path.expanduser('~/.bash_profile') if platform.system() == 'Darwin' else os.path.expanduser('~/.bashrc')
+        if bindir in os.environ['PATH']:
+            return
+        lines = []
+        with open(cfg_file) as fp:
+            lines = fp.readlines()
+        lines = [f'\nexport PATH="{bindir}:$PATH"\n\n'] + lines
+        with open(cfg_file, 'w') as fp:
+            fp.writelines(lines)
+    os.environ['PATH'] = bindir + os.pathsep + os.environ['PATH']
 
 
-class Singleton(_Singleton('SingletonMeta', (object,), {})):
-    pass
+def remove_from_os_paths(bindir):
+    if platform.system() == 'Windows':
+        import winreg
+        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as reg:
+            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Environment', 0, winreg.KEY_ALL_ACCESS) as key:
+                user_paths, _ = winreg.QueryValueEx(key, 'Path')
+                start = user_paths.find(bindir)
+                if not bindir in user_paths:
+                    return
+                import re
+                # escape to handle metachars
+                pattern = re.compile('|'.join(map(re.escape, bindir)))
+                pattern.sub('', user_paths)
+                winreg.SetValueEx(key, 'Path', 0, winreg.REG_EXPAND_SZ, user_paths)
+    else:
+        cfg_file = os.path.expanduser('~/.bash_profile') if platform.system() == 'Darwin' else os.path.expanduser('~/.bashrc')
+        if not bindir in os.environ['PATH']:
+            return
+        keyword = r'^[\s](*)export PATH="{}:$PATH"'.format(bindir)
+        # escape to handle metachars
+        pattern_prepend = f'export PATH="{bindir}:$PATH"'
+        pattern_append = f'export PATH="$PATH:{bindir}"'
+        lines = []
+        with open(cfg_file) as fp:
+            lines = fp.readlines()
+        for l, line in enumerate(lines):
+            line = line.strip()
+            if line.startswith(pattern_prepend):
+                lines[l] = line.replace(pattern_prepend, '')
+            if line.startswith(pattern_append):
+                lines[l] = line.replace(pattern_append, '')
+        with open(cfg_file, 'w') as fp:
+            fp.writelines(lines)
+    os.environ['PATH'] = os.environ['PATH'].replace(bindir, '')
 
 
-def test():
+def _test():
     pass
 
 
 if __name__ == '__main__':
-    test()
+    _test()
